/*
 * Copyright © 2012 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Thomas Voß <thomas.voss@canonical.com>
 */
#ifndef CORE_DBUS_INTERFACES_DBUS_H_
#define CORE_DBUS_INTERFACES_DBUS_H_

#include <core/dbus/bus.h>
#include <core/dbus/macros.h>
#include <core/dbus/object.h>
#include <core/dbus/signal.h>

#include <core/dbus/types/object_path.h>

#include <core/dbus/types/stl/tuple.h>
#include <core/dbus/types/stl/string.h>

namespace core
{
namespace dbus
{
namespace interfaces
{
/** @brief Provides access to the special message bus name org.freedesktop.DBus. */
struct DBus
{
    /** @brief Returns the name of the interface. */
    static inline const std::string& name()
    {
        static const std::string s{"org.freedesktop.DBus"};
        return s;
    }

    /** @brief Returns the default path under which the daemon is known. */
    static inline const core::dbus::types::ObjectPath& path()
    {
        static const core::dbus::types::ObjectPath op{"/org/freedesktop/DBus"};
        return op;
    }

    /**
      * @brief Before an application is able to send messages to other applications it
      * must send the org.freedesktop.DBus.Hello message to the message bus to
      * obtain a unique name. If an application without a unique name tries to
      * send a message to another application, or a message to the message bus
      * itself that isn't the org.freedesktop.DBus.Hello message, it will be
      * disconnected from the bus.
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(Hello, DBus, 5000)

    /**
      * @brief Gets the unique ID of the bus. The unique ID here is shared among
      * all addresses the bus daemon is listening on (TCP, UNIX domain socket,
      * etc.) and its format is described in the section called “UUIDs”. Each
      * address the bus is listening on also has its own unique ID, as described
      * in the section called “Server Addresses”. The per-bus and per-address
      * IDs are not related. There is also a per-machine ID, described in the
      * section called “org.freedesktop.DBus.Peer” and returned by
      * org.freedesktop.DBus.Peer.GetMachineId(). For a desktop session bus, the
      * bus ID can be used as a way to uniquely identify a user's session.
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(GetId, DBus, 5000)

    /**
      * @brief Returns a list of all currently-owned names on the bus.
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(ListNames, DBus, 5000)

    /**
      * @brief Returns a list of all names that can be activated on the bus.
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(ListActivatableNames, DBus, 5000)

    /**
      * @brief Checks if the specified name exists (currently has an owner).
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(NameHasOwner, DBus, 5000)

    /**
      * @brief Tries to launch the executable associated with a name.
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(StartServiceByName, DBus, 5000)

    /**
      * @brief Normally, session bus activated services inherit the environment
      * of the bus daemon. This method adds to or modifies that environment when
      * activating services.
      *
      * Some bus instances, such as the standard system bus, may disable access
      * to this method for some or all callers. Note, both the environment
      * variable names and values must be valid UTF-8. There's no way to update
      * the activation environment with data that is invalid UTF-8.
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(UpdateActivationEnvironment, DBus, 5000)

    /**
     * @brief Returns the Unix user ID of the process connected to the server.
     * If unable to determine it (for instance, because the process is not on
     * the same machine as the bus daemon), an error is returned.
     */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(GetConnectionUnixUser, DBus, 5000)

    /**
     * @brief Returns the Unix process ID of the process connected to the
     * server. If unable to determine it (for instance, because the process is
     * not on the same machine as the bus daemon), an error is returned.
     */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(GetConnectionUnixProcessID, DBus, 5000)

    /**
      * @brief The org.freedesktop.DBus.GetConnectionAppArmorSecurityContext() D-Bus method gets
      * the AppArmor confinement string associated with the unique connection name. If
      * D-Bus is not performing AppArmor mediation, the
      * org.freedesktop.DBus.Error.AppArmorSecurityContextUnknown error is returned.
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(GetConnectionAppArmorSecurityContext, DBus, 5000)

    /**
      * @brief Adds a match rule to match messages going through the message bus.
      * If the bus does not have enough resources the org.freedesktop.DBus.Error.OOM error is returned.
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(AddMatch, DBus, 5000)

    /**
      * @brief Removes the first rule that matches (see the section called “Match Rules”).
      * If the rule is not found the org.freedesktop.DBus.Error.MatchRuleNotFound error is returned.
      */
    DBUS_CPP_METHOD_WITH_TIMEOUT_DEF(RemoveMatch, DBus, 5000)

    /** @brief Summarizes all signals emitted by the dbus daemon. */
    struct Signals
    {
        /** @cond */
        Signals() = delete;
        /** @endcond */

        /** @brief The signal invocation argument for NameOwnerChanged. */
        typedef std::tuple
        <
            std::string, // Name with a new owner
            std::string, // Old owner or empty string if none
            std::string  // New owner or empty string if none
        > NameOwnerChangedTuple;

        /**
          * @brief This signal indicates that the owner of a name has changed.
          * It's also the signal to use to detect the appearance of new names on
          * the bus.
          */
        DBUS_CPP_SIGNAL_DEF(NameOwnerChanged, DBus, NameOwnerChangedTuple)

        /**
          * @brief This signal is sent to a specific application when it loses ownership of a name.
          */
        DBUS_CPP_SIGNAL_DEF(NameLost, DBus, std::string)
    };

    /** @brief Provides applications with access to the methods and signals offered by the dbus daemon. */
    struct Stub
    {

        /** @brief Constructs a new stub instance for the given object. */
        inline Stub(const Object::Ptr& object)
            : object{object},
              signals
              {
                  object->get_signal<Signals::NameLost>(),
                  object->get_signal<Signals::NameOwnerChanged>()
              }
        {
        }

        /** @brief Constructs a new stub instance for the given service instance. */
        inline Stub(const Service::Ptr& service)
            : Stub(service->object_for_path(DBus::path()))
        {
        }

        /** @brief Constructs a new stub instance for the given bus connection. */
        inline Stub(const Bus::Ptr& connection)
            : Stub(Service::use_service<DBus>(connection))
        {
        }

        /** @brief Stub object referring to the dbus daemon. */
        Object::Ptr object;

        /** @brief All signals emitted by the dbus daemon. */
        struct
        {
            /**
              * @brief This signal is sent to a specific application when it loses ownership of a name.
              */
            Signal<Signals::NameLost, Signals::NameLost::ArgumentType>::Ptr name_lost;

            /**
              * @brief This signal indicates that the owner of a name has changed.
              * It's also the signal to use to detect the appearance of new names on
              * the bus.
              */
            Signal<Signals::NameOwnerChanged, Signals::NameOwnerChanged::ArgumentType>::Ptr name_owner_changed;
        } signals;

    };
};
}
}
}

#endif // CORE_DBUS_INTERFACES_DBUS_H_



