/*
 * Copyright © 2013 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Thomas Voß <thomas.voss@canonical.com>
 */

#include <core/dbus/interfaces/dbus.h>

#include <core/dbus/fixture.h>

#include <core/dbus/asio/executor.h>

#include <core/posix/signal.h>

#include <core/testing/cross_process_sync.h>
#include <core/testing/fork_and_run.h>

#include <gtest/gtest.h>

#include "test_service.h"

namespace
{
struct OrgFreedesktopDBus : public core::dbus::testing::Fixture
{
    OrgFreedesktopDBus()
        : test_service
          {
              core::posix::fork([this]()
              {
                  auto trap = core::posix::trap_signals_for_all_subsequent_threads({core::posix::Signal::sig_term});
                  trap->signal_raised().connect([trap](core::posix::Signal)
                  {
                      trap->stop();
                  });

                  auto bus = session_bus();
                  bus->install_executor(core::dbus::asio::make_executor(bus));

                  std::thread t{[bus]() { bus->run(); }};

                  auto service = core::dbus::Service::add_service<test::Service>(bus);
                  auto skeleton = service->add_object_for_path(core::dbus::types::ObjectPath("/this/is/unlikely/to/exist/Service"));

                  trap->run();

                  bus->stop();

                  if (t.joinable())
                      t.join();

                  return core::posix::exit::Status::success;
              }, core::posix::StandardStream::empty)
          }
    {
    }
    core::posix::ChildProcess test_service;
};
}

TEST_F(OrgFreedesktopDBus, CallingOutToOrgFreedesktopDBusWorks)
{
    // We have to wait for the service to come up.
    std::this_thread::sleep_for(std::chrono::milliseconds{500});

    auto child = core::posix::fork([this]()
    {
        auto bus = session_bus();
        bus->install_executor(core::dbus::asio::make_executor(bus));

        std::thread t{[bus]() { bus->run(); }}; t.detach();

        core::dbus::interfaces::DBus::Stub stub(bus);

        auto r2 = stub.object->transact_method<core::dbus::interfaces::DBus::GetId, std::string>();
        EXPECT_FALSE(r2.is_error()); auto id = r2.value();

        auto r3 = stub.object->transact_method<core::dbus::interfaces::DBus::ListNames, std::vector<std::string>>();
        EXPECT_FALSE(r3.is_error()); auto names = r3.value();

        auto r4 = stub.object->transact_method<core::dbus::interfaces::DBus::ListActivatableNames, std::vector<std::string>>();
        EXPECT_FALSE(r4.is_error()); auto activatable_names = r4.value();

        auto r5 = stub.object->transact_method<core::dbus::interfaces::DBus::NameHasOwner, bool>(core::dbus::traits::Service<test::Service>::interface_name());
        EXPECT_FALSE(r5.is_error()); auto name_has_owner = r5.value(); EXPECT_TRUE(name_has_owner);

        auto r6 = stub.object->transact_method<core::dbus::interfaces::DBus::GetConnectionUnixUser, std::uint32_t>(core::dbus::traits::Service<test::Service>::interface_name());
        EXPECT_FALSE(r6.is_error()); auto user = r6.value(); EXPECT_EQ(::getuid(), user);

        auto r7 = stub.object->transact_method<core::dbus::interfaces::DBus::GetConnectionUnixProcessID, std::uint32_t>(core::dbus::traits::Service<test::Service>::interface_name());
        EXPECT_FALSE(r7.is_error());

        auto r8 = stub.object->transact_method<core::dbus::interfaces::DBus::GetConnectionAppArmorSecurityContext, std::string>(core::dbus::traits::Service<test::Service>::interface_name());
        EXPECT_FALSE(r8.is_error()); auto aa_profile = r8.value(); EXPECT_EQ("unconfined", aa_profile);

        auto r9 = stub.object->transact_method<core::dbus::interfaces::DBus::AddMatch, void>(core::dbus::MatchRule().sender(std::string{"org.freedesktop.DBus"}).as_string());
        EXPECT_FALSE(r9.is_error());

        auto r10 = stub.object->transact_method<core::dbus::interfaces::DBus::RemoveMatch, void>(core::dbus::MatchRule().sender(std::string{"org.freedesktop.DBus"}).as_string());
        EXPECT_FALSE(r10.is_error());

        bus->stop();

        return ::testing::Test::HasFailure() ? core::posix::exit::Status::failure : core::posix::exit::Status::success;
    }, core::posix::StandardStream::empty);

    auto result = child.wait_for(core::posix::wait::Flags::untraced);

    EXPECT_EQ(core::posix::wait::Result::Status::exited, result.status);
    EXPECT_EQ(core::posix::exit::Status::success, result.detail.if_exited.status);
}

TEST_F(OrgFreedesktopDBus, SignalsAreReceivedCorrectly)
{
    // We have to wait for the service to come up.
    std::this_thread::sleep_for(std::chrono::milliseconds{500});

    auto child = core::posix::fork([this]()
    {
        auto trap = core::posix::trap_signals_for_all_subsequent_threads({core::posix::Signal::sig_term});
        trap->signal_raised().connect([trap](core::posix::Signal)
        {
            trap->stop();
        });

        auto bus = session_bus();
        bus->install_executor(core::dbus::asio::make_executor(bus));

        std::thread t{[bus]() { bus->run(); }}; t.detach();

        core::dbus::interfaces::DBus::Stub stub(bus);

        stub.signals.name_owner_changed->connect([trap](core::dbus::interfaces::DBus::Signals::NameOwnerChangedTuple tuple)
        {
            if (std::get<0>(tuple) == core::dbus::traits::Service<test::Service>::interface_name())
                trap->stop();
        });

        trap->run();

        bus->stop();

        return ::testing::Test::HasFailure() ? core::posix::exit::Status::failure : core::posix::exit::Status::success;
    }, core::posix::StandardStream::empty);

    // We have to wait for the service to come up.
    std::this_thread::sleep_for(std::chrono::milliseconds{500});

    // We forcefully shut down the test service, thus provoking a name-owner changed signal.
    test_service.send_signal_or_throw(core::posix::Signal::sig_kill);
    test_service.wait_for(core::posix::wait::Flags::untraced);

    auto result = child.wait_for(core::posix::wait::Flags::untraced);

    EXPECT_EQ(core::posix::wait::Result::Status::exited, result.status);
    EXPECT_EQ(core::posix::exit::Status::success, result.detail.if_exited.status);
}
